#!/bin/env python3

from enum import IntEnum
from typing import List
import numpy as np
from .product_data import ProductData, ProductDataPolar, ProductDataRect, ProductDataVertLevels
from .lzw_c import lzw_expand, lzw_compress

class ProductTable:
    """
    Represents a product table with a name, size, and associated data.

    Attributes:
        _name (str): The name of the table.
        _size (int): The size (number of bytes) of the table.
        _data (np.ndarray): The table data stored as a NumPy array.
    """
    def __init__(self, name: str=None, size: int = 0, data=None) -> None:
        """
        Initializes a new ProductTable.

        Args:
            name (str, optional): The name of the table. Defaults to None.
            size (int, optional): The size of the table. Defaults to 0.
            data: The data associated with the table. Defaults to None.
        """
        self._name: str = name
        self._size: int = size
        self._data: np.ndarray = data

    @property
    def name(self) -> str:
        """
        Gets the name of the product table.

        Returns:
            str: The name of the table.
        """
        return self._name
    
    @name.setter
    def name(self, value: str) -> None:
        """
        Sets the name of the product table.

        Args:
            value (str): The new table name.
        """
        self._name = value
    
    @property
    def size(self) -> int:
        """
        Gets the size of the product table.

        Returns:
            int: The size of the table.
        """
        return self._size
    
    @size.setter
    def size(self, value: int) -> None:
        """
        Sets the size of the product table.

        Args:
            value (int): The new size of the table.
        """
        self._size = value
    
    @property
    def data(self) -> np.ndarray:
        """
        Gets the data stored in the product table.

        Returns:
            np.ndarray: The table data.
        """
        return self._data
    
    @data.setter
    def data(self, value) -> None:
        """
        Sets the data for the product table.

        Args:
            value: The new data to assign to the table.
        """
        self._data = value
        
class ProductDataType(IntEnum):
    """
    Enumeration of product data types.

    Attributes:
        Unknown (int): Unknown data type.
        Polar (int): Polar product data.
        Rect (int): Rectangular product data.
        VertLevels (int): Vertical levels product data.
        # Additional types (e.g., Vad, Vvp, Rhi, etc.) can be added.
    """
    Unknown = 0
    Polar = 1
    Rect = 2
    VertLevels = 3
    #Vad, Vvp, Rhi, etc...

class ProductFile:
    """
    Represents a product file that contains header information, tables, and product data.

    Attributes:
        _data_type (ProductDataType): The type of product data.
        _data (ProductData): The loaded product data.
        _file_name (str): The name of the file.
        _header_info (list): A list of header key-value pairs.
        _tables (List[ProductTable]): A list of product tables.
        _error_descr: Error description if any error occurs.
    """
    def __init__(self) -> None:
        """
        Initializes a new ProductFile with default values.
        """
        self._data_type: ProductDataType = ProductDataType.Unknown
        self._data: ProductData = None
        self._file_name: str = None
        self._header_info: list = []
        self._tables: List[ProductTable] = []
        self._error_descr = None

    @property
    def data_type(self) -> ProductDataType:
        """
        Gets the product data type.

        Returns:
            ProductDataType: The type of product data.
        """
        return self._data_type
    
    @property
    def data(self) -> ProductData:
        """
        Gets the loaded product data.

        Returns:
            ProductData: The product data.
        """
        return self._data

    @data.setter
    def data(self, value: ProductData) -> None:
        """
        Sets the product data.

        Args:
            value (ProductData): The new product data.
        """
        self._data = value
    
    def load_file(self, file_name: str) -> None:
        """
        Loads a product file from disk, reads its header, tables, and data, and sets the product data type.

        The method reads lines until 'end_header' is encountered, stores header key-value pairs, 
        determines the product data type based on header information, reads any product tables, 
        decompresses data if necessary, reshapes it, and stores it internally.

        Args:
            file_name (str): The path to the product file.

        Raises:
            IOError: If an unexpected EOF occurs while reading the header.
        """
        self._file_name = file_name

        #cleanup internal data
        self._data_type = ProductDataType.Unknown
        self._header_info.clear()
        self._tables.clear()
        self._data = None

        #read product line by line until line 'end_header' is reached
        line: str = ""
        f = open(self._file_name, "rb")
        try:
            while line != "end_header":
                line = f.readline().decode("utf-8")
                line = line.rstrip()
                index = line.find("=")
                if index != -1:
                    name = line[0:index]
                    value = line[index+1:]

                    #search in _header_info if this key already exists
                    #if it exists saerch in a loop if there is a key with the name
                    #adding an incremental counter at the end of the name.
                    #example: 'table_name' is read. Does 'table_name' exist? If it doesn't
                    #add it to _header_info with the name 'table_name'. If later another
                    #'table_name' is read from the file we search in _header_info a key with
                    #'the name 'table_name2', if it doesn't exist, add it as 'table_name2' as name.
                    #then we go forward with this logic with 'table_name3', 'table_name4' and so on...
                    key = self.__get_valid_key_name(name)
                    self._header_info.append([key, value])
        except EOFError as ex:
            f.close()
            raise IOError("unexpected eof found while reading text header")
        
        #set product data type
        if self.__is_vad() or self.__is_vvp() or self.__is_vpr():
            self._data_type = ProductDataType.VertLevels
        else:
            format = self.find_header_info_value("format")
            if format == "RECT" or format == "STORM":
                self._data_type = ProductDataType.Rect
            elif format == "POLAR":
                self._data_type = ProductDataType.Polar

        #read product tables
        num_tables: int = int(self.find_header_info_value("table_num"))
        for count in range(1, num_tables + 1):
            key_table_name: str = "table_name" if count == 1 else "table_name" + str(count)
            key_table_size: str = "table_size" if count == 1 else "table_size" + str(count)
            table_name: str = self.find_header_info_value(key_table_name)
            table_size: int = int(self.find_header_info_value(key_table_size))

            #create a new empty product table inside _tables
            table = ProductTable(table_name, table_size)

            #read data from file into table
            table.data = np.fromfile(f, np.uint8, table.size)

            #add table to list of tables
            self._tables.append(table)

        #read data
        zip_size: int = int(self.find_header_info_value("compressed_bytes"))
        unzip_size: int = int(self.find_header_info_value("uncompressed_bytes"))
        buff_size: int = zip_size if zip_size != 0 else unzip_size

        #uncompress if necessary
        if zip_size != 0 and zip_size != unzip_size:
            buffer = np.fromfile(f, np.uint8, zip_size)
            buffer = lzw_expand(buffer, zip_size, unzip_size)
        else:
            buffer = np.fromfile(f, np.uint8, buff_size)

        #close the file
        f.close()

        #reshape data read into a 2d matrix and store inside internal _data
        if self._data_type == ProductDataType.Polar:
            num_rays: int = int(self.find_header_info_value("row"))
            num_gates: int = int(self.find_header_info_value("column"))
            data_bits: int = int(self.find_header_info_value("data_bits"))
            #create self._data as a ProductDataPolar object, here
            #buffer is a 1D array but inside the constructor of ProductDataPolar
            #it will be reshaped to be a 2D array with num_rays rows and num_gates cols
            self._data = ProductDataPolar(num_rays, num_gates, buffer, data_bits // 8)
        elif self._data_type == ProductDataType.Rect:
            x: int = int(self.find_header_info_value("column"))
            y: int = int(self.find_header_info_value("row"))
            xres: float = float(self.find_header_info_value("rect_xres"))
            yres: float = float(self.find_header_info_value("rect_yres"))
            data_bits: int = int(self.find_header_info_value("data_bits"))
            #create self._data as a ProductDataRect object, here
            #buffer is a 1D array but inside the constructor of ProductDataRect
            #it will be reshaped to be a 2D array with y rows and x cols
            self._data = ProductDataRect(x, y, xres, yres, data_bits // 8, buffer)
        elif self._data_type == ProductDataType.VertLevels:
            num_floats32: int = int(self.find_header_info_value("row"))
            num_levels: int = int(self.find_header_info_value("column"))
            #create self._data as a ProductDataVertLevels
            self._data = ProductDataVertLevels(num_floats32, num_levels, buffer)

    def save_file(self, file_name: str, compress: bool=True) -> None:
        """
        Saves the product file to disk with header information, tables, and product data.

        If compression is enabled, data is compressed using LZW before saving.

        Args:
            file_name (str): The path where the product file will be saved.
            compress (bool, optional): Flag to indicate if data should be compressed. Defaults to True.

        Raises:
            IOError: If required header information for compression is missing or if compression fails.
        """
        #compressed_bytes and uncompressed_bytes parameters must be present
        info_compressed = self.find_header_info("compressed_bytes")
        if info_compressed is None:
            raise IOError("can't find compressed_bytes info in header")
        info_uncompressed = self.find_header_info("uncompressed_bytes")
        if info_uncompressed is None:
            raise IOError("can't find info_uncompressed info in header")
        
        #compress data if compression is enabled
        buff_size: int = self._data.num_rows * self._data.num_cols * self._data.data_bytes
        assert buff_size == self._data.data.nbytes, "data size in bytes does not match the expected size calculated from num_rows, num_cols and data_bytes"
        zip_buff = None
        zip_size = 0
        if compress:
            zip_buff = lzw_compress(self._data.data.flatten(), self._data.data.nbytes)
            zip_size = zip_buff.size
            if zip_size <= 0 or zip_size > buff_size:
                raise IOError("error during compression of product data")
            info_compressed[1] = str(zip_size)
        else:
            info_compressed[1] = "0"
        info_uncompressed[1] = str(buff_size)

        #openm file for binary writing
        self._file_name = file_name
        f = open(self._file_name, "wb")

        #save header info
        for name, value in self._header_info:
            key: str = self.__normalize_key_name_for_save(name)
            line: str = "%s=%s\n" % (key, value)
            f.write(line.encode())
        f.write("end_header\n".encode())

        #save tables
        for table in self._tables:
            table.data.tofile(f)

        #save data
        if compress and zip_buff is not None and zip_size > 0:
            f.write(zip_buff)
        else:
            self._data.data.tofile(f)

        f.close()

    def get_header_info(self) -> list:
        """
        Returns the header information as a list of key-value pairs.

        Returns:
            list: The header information.
        """
        return self._header_info

    def get_tables(self) -> list:
        """
        Returns the list of product tables.

        Returns:
            list: List of ProductTable objects.
        """
        return self._tables

    def get_table(self, table_name: str):
        """
        Retrieves a product table by its name.

        Args:
            table_name (str): The name of the table to retrieve.

        Returns:
            ProductTable: The matching table if found; otherwise, None.
        """
        ret_table = [x for x in self._tables if x.name == table_name]
        return ret_table[0] if len(ret_table) > 0 else None
    
    def add_table(self, table_name: str) -> ProductTable:
        """
        Adds a new empty product table with the given name.

        Args:
            table_name (str): The name of the new table.

        Returns:
            ProductTable: The newly created product table.
        """
        new_table = ProductTable(table_name)
        self._tables.append(new_table)
        return new_table

    def add_header_info(self, name: str, value: str):
        """
        Adds a new header info key-value pair.

        Args:
            name (str): The header info key.
            value (str): The header info value.

        Raises:
            ValueError: If a header info with the same name already exists.
        """
        info = self.find_header_info(name)
        if info is not None:
            raise ValueError("can't add header info '%s': an info with that name is already present" % name)
        
        self._header_info.append([name, value])

    def find_header_info(self, search: str):
        """
        Searches for a header info key-value pair by key.

        Args:
            search (str): The key to search for.

        Returns:
            list: A list containing the key and value if found; otherwise, None.
        """
        list = [x for x in self._header_info if x[0] == search]
        return list[0] if len(list) > 0 else None

    def find_header_info_value(self, search: str) -> str:
        """
        Retrieves the value for a given header info key.

        Args:
            search (str): The header info key.

        Returns:
            str: The corresponding value, or None if not found.
        """
        info = self.find_header_info(search)
        return None if info is None else info[1]

    def find_header_param_value(self, search: str):
        """
        Searches for a header parameter value by its name.

        It retrieves the total number of parameters from the header and iterates through
        them to find a matching parameter name.

        Args:
            search (str): The parameter name to search for.

        Returns:
            str: The parameter value if found; otherwise, None.
        """
        #get number of params
        info = self.find_header_info("param_num")
        if info is None:
            return None
        
        num_params: int =  int(info[1])
        #iterate all the params searching for 'search'
        for i in range(1, num_params + 1):
            key_param_name: str = "param_name" if i == 1 else "param_name" + str(i)
            param_name: str = self.find_header_info_value(key_param_name)
            if param_name == search:
                key_param_value: str = "param_value" if i == 1 else "param_value" + str(i)
                return self.find_header_info_value(key_param_value)

    def get_file_name(self):
        """
        Returns the name of the currently loaded file.

        Returns:
            str: The file name.
        """
        return self._file_name
    
    def is_8bit_uint(self) -> bool:
        """
        Determines if the product data is of type 8-bit unsigned integer.

        Returns:
            bool: True if the data type is 8-bit unsigned integer, False otherwise.
        """
        data_type = self.find_header_info_value("data_type")
        data_bits = self.find_header_info_value("data_bits")

        type_is_byte = data_type is not None and data_type.lower() == "byte"
        bits_is_8    = data_bits is not None and data_bits.lower() == "8"

        # True if the available information (even partial) is consistent with 8-bit unsigned int.
        return type_is_byte or bits_is_8

    def is_32bit_float(self) -> bool:
        """
        Determines if the product data is of type 32-bit float.

        Returns:
            bool: True if the data type is 32-bit float, False otherwise.
        """
        data_type = self.find_header_info_value("data_type")
        data_bits = self.find_header_info_value("data_bits")

        type_is_float = data_type is not None and data_type.lower() == "float"
        bits_is_32    = data_bits is not None and data_bits.lower() == "32"

        # True if the available information (even partial) is consistent with 8-bit unsigned int.
        return type_is_float or bits_is_32

    def get_conv_table(self) -> np.ndarray:
        """
        Retrieves the conversion table data as a NumPy array if available.

        The method checks for the presence of a conversion table in the product tables,
        determines its endianness based on header information, and returns the table data
        as a NumPy array with the appropriate format.

        Returns:
            np.ndarray: The conversion table data if found; otherwise, None.
        """
        #try to detect endianness of data
        big_endian = False
        table_endianness = self.find_header_info_value("table_endianness")
        if table_endianness is not None:
            table_endianness_lower = table_endianness.lower()
            if "big" in table_endianness_lower:
                big_endian = True

        #try to search for a conversion table in the product tables
        conv_table = None
        for table in self.get_tables():
            table_name_lower = table.name.lower()
            if "8bit" in table_name_lower and "metranet" in table_name_lower:
                conv_table = table
                break

        format = ">f4" if big_endian else "<f4"
        return conv_table.data.view(format) if conv_table is not None else None
    
    def __get_valid_key_name(self, key: str):
        """
        Generates a valid unique key name for header info by appending an incremental counter if needed.

        Args:
            key (str): The original key name.

        Returns:
            str: A unique key name that does not already exist in the header info.
        """
        count: int = 1
        while True:
            search: str = key if count == 1 else key + str(count)
            if len([x for x in self._header_info if x[0] == search]) == 0:
                return search
            count += 1

    def __normalize_key_name_for_save(self, key: str):
        """
        Normalizes a header info key name for saving, ensuring common keys are standardized.

        Args:
            key (str): The key name to normalize.

        Returns:
            str: The normalized key name.
        """
        keys_list = ["table_name", "table_size", "param_name", "param_value"]
        for item in keys_list:
            if key.startswith(item):
                return item
            
        return key
    
    def __is_vad(self):
        """
        Determines if the product file corresponds to a VAD product.

        Returns:
            bool: True if it is a VAD product, False otherwise.
        """
        pid = self.find_header_info_value("pid")
        if pid[0] == 'V' and pid[1] == 'A':
            return True
        
        format = self.find_header_info_value("format")
        if format == "VADSTR":
            return True
        
        return False

    def __is_vvp(self):
        """
        Determines if the product file corresponds to a VVP product.

        Returns:
            bool: True if it is a VVP product, False otherwise.
        """
        pid = self.find_header_info_value("pid")
        if pid[0] == 'V' and pid[1] == 'V':
            return True
        
        format = self.find_header_info_value("format")
        if format == "VVPSTR":
            return True
        
        return False

    def __is_vpr(self):
        """
        Determines if the product file corresponds to a VPR product.

        Returns:
            bool: True if it is a VPR product, False otherwise.
        """
        pid = self.find_header_info_value("pid")
        if pid[0] == 'Z' and pid[1] == 'Z':
            return True
        
        return False


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2026, Eldes, Matteo Galletti <m.galletti@eldes.it>

import glob
from setuptools import setup, find_packages

requirements = [
    'numpy',
]

setup_requirements = [
#    'setuptools_scm',
]

test_requirements = [
]

extras = {
}

packages = find_packages(include=['pymetranet'])

package_dir = {}

package_data = {}

scripts = glob.glob('examples/*.py') 

setup(
    name='pymetranet',
    #use_scm_version=True,
    version="0.4.0",
    author="Eldes",
    author_email='info@eldes.t',
    description='Python Metranet library',
    url='https://www.eldesradar.com',
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
    ],
    keywords='pymetranet',
    entry_points={},
    scripts=scripts,
    license="BSD-3-Clause license",
    long_description=open('README.md').read(),
    include_package_data=True,
    zip_safe=False,
    test_suite='test',
    py_modules=['pymetranet'],
    packages=packages,
    install_requires=requirements,
    package_dir=package_dir,
    package_data=package_data,
    setup_requires=setup_requirements,
    tests_require=test_requirements,
    extras_require=extras
)



#!/bin/env python

#import from system
import sys
import os
import math
import argparse
import time
import numpy as np
from PIL import Image, ImageDraw

#safe pymetranet import
import import_pymetranet
from pymetranet import PolarSweepSerializer, PolarSweep, PolarPpiData

ERADIUS = 6371000.0 * 4.0 / 3.0
EARTH_RADIUS = ERADIUS * 0.001
EARTH_RADIUS2 = EARTH_RADIUS * EARTH_RADIUS
    
def calc_range(gate_range, sin_elev, cos_elev, radar_height):
    global EARTH_RADIUS, EARTH_RADIUS2
    
    gate_height = radar_height + math.sqrt(EARTH_RADIUS2 +
        gate_range * gate_range +
        2 * EARTH_RADIUS * gate_range * sin_elev) - EARTH_RADIUS
    gate_horizon_distance = EARTH_RADIUS * math.asin(gate_range * cos_elev / (EARTH_RADIUS + gate_height))
    
    return gate_height, gate_horizon_distance
    
def generate_waterfall_image_from_polar(polar: PolarPpiData):
    img_waterfall = Image.new("RGB", (polar.num_gates, polar.num_rays), (255, 255, 255))
    for i in range(polar.num_rays):
        for j in range(polar.num_gates):
            if not math.isnan(polar.data[i][j]):
                img_waterfall.putpixel((j, i), (0, 0, 0))

    return img_waterfall

def generate_rect_image_from_rect(rect: np.ndarray):
    height = rect.shape[0]
    width = rect.shape[1]
    img_rect = Image.new("RGB", (width, height), (255, 255, 255))
    for i in range(height):
        for j in range(width):
            if not math.isnan(rect[i][j]):
                img_rect.putpixel((j, i), (0, 0, 0))

    return img_rect

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--mom", type=str, default="Z")
    parser.add_argument('infile', nargs='*', type=str)
    
    args = parser.parse_args()
    
    for file_name in args.infile:
        print("loading file %s..." % file_name)
        t_start = time.perf_counter()
        loaded_sweep = PolarSweepSerializer.load(file_name)
        t_stop = time.perf_counter()
        print("file %s successfully loaded in %.3f secs." % (file_name, t_stop-t_start))
        
        #transform 'mom'
        t_start = time.perf_counter()
        polar = PolarPpiData()
        polar.transform(loaded_sweep, mom_name=args.mom)
        t_stop = time.perf_counter()
        print("moment '%s' transformed into polar from loaded sweep in in %.3f secs." % (args.mom, t_stop-t_start))
        #print("polar data rays: %d gates: %d" % (polar.num_rays, polar.num_gates))
        #for ray_row in range(polar.num_rays):
            #for gate_col in range(polar.num_gates):
                #if not math.isnan(polar.data[ray_row][gate_col]):
                    #print("data[%03d][%03d]: %g" % (ray_row, gate_col, polar.data[ray_row][gate_col]))
        
        #generate waterfall image from polar object
        t_start = time.perf_counter()
        img_waterfall = generate_waterfall_image_from_polar(polar)
        t_stop = time.perf_counter()
        print("waterfall image from polar generated in %.3f secs." % (t_stop-t_start))
        
        #save waterfall image to disk
        t_start = time.perf_counter()
        script_dir = os.path.dirname(os.path.abspath(sys.argv[0]))
        img_waterfall_path = os.path.join(script_dir, os.path.basename(file_name) + ".waterfall.png")
        img_waterfall.save(img_waterfall_path)
        t_stop = time.perf_counter()
        print("waterfall image saved in %s in %.3f secs." % (img_waterfall_path, t_stop-t_start))

        #generate rect from polar
        t_start = time.perf_counter()
        gw = loaded_sweep.sweepheader.gatewidth
        rect = polar.polar2rect(gw, vectorized=False)
        t_stop = time.perf_counter()
        print("polar2rect generated in %.3f secs." % (t_stop-t_start))

        #generate rect image from rect object
        t_start = time.perf_counter()
        img_rect = generate_rect_image_from_rect(rect)
        t_stop = time.perf_counter()
        print("rect image from rect generated in %.3f secs." % (t_stop-t_start))

        #save rect image to disk
        t_start = time.perf_counter()
        img_rect_path = os.path.join(script_dir, os.path.basename(file_name) + ".rect.png")
        img_rect.save(img_rect_path)
        t_stop = time.perf_counter()
        print("rect image saved in %s in %.3f secs." % (img_rect_path, t_stop-t_start))

        #generate rect from polar using vectorized technique
        t_start = time.perf_counter()
        gw = loaded_sweep.sweepheader.gatewidth
        rect_vect = polar.polar2rect(gw, size=None, vectorized=True)
        t_stop = time.perf_counter()
        print("polar2rect vectorized generated in %.3f secs." % (t_stop-t_start))

         #generate rect image from rect_vect object
        t_start = time.perf_counter()
        img_rect_vect = generate_rect_image_from_rect(rect_vect)
        t_stop = time.perf_counter()
        print("rect_vect image from rect_vect generated in %.3f secs." % (t_stop-t_start))

        #save rect_vect image to disk
        t_start = time.perf_counter()
        img_rect_vect_path = os.path.join(script_dir, os.path.basename(file_name) + ".rect_vect.png")
        img_rect_vect.save(img_rect_vect_path)
        t_stop = time.perf_counter()
        print("rect_vect image saved in %s in %.3f secs." % (img_rect_vect_path, t_stop-t_start))

    return 0

if __name__ == '__main__':
    main()



#!/bin/env python3

import struct
from .volumesweep import SweepHeader, MomentInfo
from .volumesweep import  RayHeader
from .msx_serializer import MSxSerializer

class MSxV1Serializer(MSxSerializer):
    def __init__(self):
        super().__init__()
        
    def read_sweep_header(self, f):
        ret_sweepheader = SweepHeader()
        
        fmt = "=4sB3BI16s16sfffBBBBBB2BHHffffI"
        struct_len = struct.calcsize(fmt)
        data = f.read(struct_len)
        if not data:
            f.close()
            self.eof = True
            raise Exception("Error reading sweep header")
        s = struct.Struct(fmt)
        unpacked_data = s.unpack(data)
        
        ret_sweepheader.fileid = MSxSerializer.stringify(unpacked_data[0])
        ret_sweepheader.version = unpacked_data[1]
        #unpacked_data[2], 3 and 4 are spare
        ret_sweepheader.length = unpacked_data[5]
        ret_sweepheader.radarname = MSxSerializer.stringify(unpacked_data[6])
        ret_sweepheader.scanname = MSxSerializer.stringify(unpacked_data[7])
        ret_sweepheader.radarlat = unpacked_data[8]
        ret_sweepheader.radarlon = unpacked_data[9]
        ret_sweepheader.radarheight = unpacked_data[10]
        ret_sweepheader.sequencesweep = unpacked_data[11]
        ret_sweepheader.currentsweep = unpacked_data[12]
        ret_sweepheader.totalsweep = unpacked_data[13]
        ret_sweepheader.antmode = unpacked_data[14]
        ret_sweepheader.priority = unpacked_data[15]
        ret_sweepheader.quality = unpacked_data[16]
        #unpacked_data[17] and 18 are spare
        ret_sweepheader.repeattime = unpacked_data[19]
        ret_sweepheader.nummoments = unpacked_data[20]
        ret_sweepheader.gatewidth = unpacked_data[21]
        ret_sweepheader.wavelength = unpacked_data[22]
        ret_sweepheader.pulsewidth = unpacked_data[23]
        ret_sweepheader.startrange = unpacked_data[24]
        ret_sweepheader.metadatasize = unpacked_data[25]
        
        #read moments information
        fmt = "=IBB2B12s12sfffB3B"
        struct_len = struct.calcsize(fmt)
        s = struct.Struct(fmt)
        for i in range(ret_sweepheader.nummoments):
            data = f.read(struct_len)
            if not data:
                f.close()
                self.eof = True
                raise Exception("Error reading moment info structure")
            unpacked_data = s.unpack(data)
            
            mom_info = MomentInfo()
            mom_info.momentid = unpacked_data[0]
            mom_info.dataformat = unpacked_data[1]
            mom_info.numbytes = unpacked_data[2]
            #unpacked_data[3] and 4 are spare
            mom_info.name = MSxSerializer.stringify(unpacked_data[5])
            mom_info.unit = MSxSerializer.stringify(unpacked_data[6])
            mom_info.factora = unpacked_data[7]
            mom_info.factorb = unpacked_data[8]
            mom_info.factorc = unpacked_data[9]
            mom_info.scaletype = unpacked_data[10]
            #unpacked_data[11], 12 and 13 are spare
            
            ret_sweepheader.momentsinfo.append(mom_info)
        
        #read sweep metadata
        if ret_sweepheader.metadatasize > 0:
            fmt = "=%ds" % ret_sweepheader.metadatasize
            struct_len = struct.calcsize(fmt)
            s = struct.Struct(fmt)
            data = f.read(struct_len)
            if not data:
                f.close()
                self.eof = True
                raise Exception("Error reading sweep header metadata")
            unpacked_data = s.unpack(data)
            
            ret_sweepheader.metadata = MSxSerializer.stringify(unpacked_data[0])
            
        return ret_sweepheader
        
    def read_ray_header(self, f):
        ret_rayheader = RayHeader()
        
        fmt = "=IIIHHIfQII"
        struct_len = struct.calcsize(fmt)
        data = f.read(struct_len)
        if not data:
            self.eof = True
            return None
        s = struct.Struct(fmt)
        unpacked_data = s.unpack(data)
        
        ret_rayheader.length = unpacked_data[0]
        ret_rayheader.startangle = unpacked_data[1]
        ret_rayheader.endangle = unpacked_data[2]
        ret_rayheader.sequence = unpacked_data[3]
        ret_rayheader.numpulses = unpacked_data[4]
        ret_rayheader.databytes = unpacked_data[5]
        ret_rayheader.prf = unpacked_data[6]
        ret_rayheader.datetime = unpacked_data[7]
        ret_rayheader.dataflags = unpacked_data[8]
        ret_rayheader.metadatasize = unpacked_data[9]
        
        #read ray metadata
        if ret_rayheader.metadatasize > 0:
            fmt = "=%ds" % ret_rayheader.metadatasize
            struct_len = struct.calcsize(fmt)
            s = struct.Struct(fmt)
            data = f.read(struct_len)
            if not data:
                f.close()
                self.eof = True
                raise Exception("Error reading ray header metadata")
            unpacked_data = s.unpack(data)
            
            ret_rayheader.metadata = MSxSerializer.stringify(unpacked_data[0])
        
        return ret_rayheader


#!/bin/env python3

from enum import IntEnum
from abc import ABC, abstractmethod
from typing import Generic, TypeVar, List, Dict
import xml.etree.ElementTree as ET
from xml.dom import minidom
import re

class CmdLineParams:
    def __init__(self):
        self._cmdline = ""
        self._params = {}
        
    @property
    def params(self) -> Dict[str, str]:
        return self._params
        
    def parse_cmd_line(self, *args):
        """
        Parse the command line.

        This method can be called in two ways:
        1. With argc and argv: parse_cmd_line(argc, argv)
        2. With a single string: parse_cmd_line(cmdLine)

        Args:
            *args: Either (argc, argv) or (cmdLine,)

        Returns:
            int: The number of parameters processed or a value less than
                 zero in case of an error.
        """
        if len(args) == 2:
            argc, argv = args
            
            cmd_line = ''
            for i in range(argc):
                #embrace the param value with double quotes at the beginning
                #and at the end so that argument 'c=3 4 5' become 'c="3 4 5"'
                arg = argv[i]
                index = arg.find('=')
                if index != -1:
                    arg = arg[:index + 1] + '"' + arg[index + 1:] + '"'
                
                #add argument to the cmdLine
                cmd_line += arg
                
                #if this is not the last element, add a space before continuing the loop
                if i + 1 < argc:
                    cmd_line += ' '
                    
            self.cmdline = cmd_line
        elif len(args) == 1:
            self.cmdline = args[0]
        else:
            return -1

        self.params.clear()
        
        if not self.cmdline:
            return -1
            
        start_index = 0
        while True:
            #find ' ' (space) starting from startIndex
            sep_index = self.cmdline.find(' ', start_index)
            if sep_index == -1:
                break
                
            #find '=' starting from sepIndex+1
            eq_index = self.cmdline.find('=', sep_index + 1)
            if eq_index == -1:
                break
                
            #find first \" starting from eqIndex+1
            pos_start = self.cmdline.find('"', eq_index + 1)
            if pos_start == -1:
                break
                
            #find next \" starting from posStart+1
            pos_end = self.cmdline.find('"', pos_start + 1)
            if pos_end == -1:
                break
                
            #get the parameter name and value
            param_name = self.cmdline[sep_index + 1:eq_index]
            param_value = self.cmdline[pos_start + 1:pos_end]
            
            #add the parameter to the vector
            self.params[param_name] = param_value
            
            #update startIndex to search for next parameter
            start_index = pos_end
            
        return len(self.params)

    def get_param_value(self, param_name):
        return self.params.get(param_name, "")

    def get_params_count(self):
        return len(self.params)
        

class ProductParamType(IntEnum):
    Undefined = 0
    Pid = 1
    Hidden = 2
    Bool = 3
    Int = 4
    Float = 5
    String = 6
    Enum = 7
    MapSizeRect = 8
    PathDir = 9
    PathFile = 10
    CatchmentFile = 11
    RadarCfg = 12
    RadarNetwork = 13


class MapSizeRect:
    def __init__(self, x_size: int=0, y_size: int=0, x_res: float=0.0, y_res: float=0.0):
        self._x_size = x_size
        self._y_size = y_size
        self._x_res = x_res
        self._y_res = y_res

    def __repr__(self) -> str:
        return f"{self.x_size};{self.y_size};{self.x_res};{self.y_res}"
        
    @property
    def x_size(self) -> int:
        return self._x_size
        
    @x_size.setter
    def x_size(self, value: int) -> None:
        self._x_size = value
        
    @property
    def y_size(self) -> int:
        return self._y_size
        
    @y_size.setter
    def y_size(self, value: int) -> None:
        self._y_size = value
        
    @property
    def x_res(self) -> float:
        return self._x_res
        
    @x_res.setter
    def x_res(self, value: float) -> None:
        self._x_res = value
        
    @property
    def y_res(self) -> float:
        return self._y_res
        
    @y_res.setter
    def y_res(self, value: float) -> None:
        self._y_res = value
        
        
class ParamBoolConverter:
    @staticmethod
    def to_string(value: bool) -> str:
        return '1' if value else '0'

    @staticmethod
    def from_string(value: str) -> bool:
        return value == '1'

        
class ParamIntConverter:
    @staticmethod
    def to_string(value: int) -> str:
        return str(value)

    @staticmethod
    def from_string(value: str) -> int:
        try:
            return int(value)
        except ValueError:
            return 0


class ParamFloatConverter:
    @staticmethod
    def to_string(value: float) -> str:
        return str(value)

    @staticmethod
    def from_string(value: str) -> float:
        try:
            return float(value)
        except ValueError:
            return 0.0


class ParamStringConverter:
    @staticmethod
    def to_string(value: str) -> str:
        return value

    @staticmethod
    def from_string(value: str) -> str:
        return value


class ParamMapSizeRectConverter:
    @staticmethod
    def to_string(value: MapSizeRect) -> str:
        return f"{value.x_size};{value.y_size};{value.x_res};{value.y_res}"

    @staticmethod
    def from_string(value: str) -> MapSizeRect:
        pos_start = 0
        pos_end = value.find(';', pos_start)
        x_size = ParamIntConverter.from_string(value[:pos_end]) if pos_end != -1 else 0
        pos_start = pos_end + 1
        pos_end = value.find(';', pos_start)
        y_size = ParamIntConverter.from_string(value[pos_start:pos_end]) if pos_end != -1 else 0
        pos_start = pos_end + 1
        pos_end = value.find(';', pos_start)
        x_res = ParamFloatConverter.from_string(value[pos_start:pos_end]) if pos_end != -1 else 0.0
        pos_start = pos_end + 1
        y_res = ParamFloatConverter.from_string(value[pos_start:]) if pos_end != -1 else 0.0
        
        return MapSizeRect(x_size, y_size, x_res, y_res)


class ProductParamGroup:
    def __init__(self, name: str=None):
        self._name = name
        
    @property
    def name(self) -> str:
        return self._name
        
    @name.setter
    def name(self, value: str) -> None:
        self._name = value
        
class ProductParam(ABC):
    def __init__(self, name: str, descr: str, param_type: ProductParamType, group: ProductParamGroup):
        self._name = name
        self._descr = descr
        self._param_type = param_type
        self._group = group
        
    @property
    def name(self) -> str:
        return self._name
        
    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @property
    def descr(self) -> str:
        return self._descr
        
    @descr.setter
    def descr(self, value: str) -> None:
        self._descr = value

    @property
    def param_type(self) -> ProductParamType:
        return self._param_type
        
    @param_type.setter
    def param_type(self, value: ProductParamType) -> None:
        self._param_type = value

    @property
    def group(self) -> ProductParamGroup:
        return self._group
        
    @group.setter
    def group(self, value: ProductParamGroup) -> None:
        self._group = value

    @abstractmethod
    def get_string_value(self) -> str:
        """
        Subclasses must implement this method
        """
        pass
        
    @abstractmethod
    def set_string_value(self, value: str) -> None:
        """
        Subclasses must implement this method
        """
        pass


T = TypeVar('T')
class ProductParamDefault(Generic[T]):
    def __init__(self, has_default: bool, default_value: T = None):
        if has_default:
            self._has_default = True
            self._default_value = default_value
        else:
            self.has_default = False
            self.default_value = None
            
    @property
    def has_default(self) -> bool:
        return self._has_default
        
    @has_default.setter
    def has_default(self, value: bool) -> None:
        self._default_value = value

    @property
    def default_value(self) -> T:
        return self._default_value

    @default_value.setter
    def default_value(self, value: T) -> None:
        self._default_value = value


class ProductParamRange(Generic[T]):
    def __init__(self, has_range: bool, min_val: T = None, max_val: T = None):
        if has_range:
            self._has_range = True
            self._min_val = min_val
            self._max_val = max_val
        else:
            self._has_range = False
            self._min_val = None
            self._max_val = None

    @property
    def has_range(self) -> bool:
        return self._has_range

    @property
    def min_val(self) -> T:
        return self._min_val

    @property
    def max_val(self) -> T:
        return self._max_val


class ProductParamHidden(ProductParam):
    def __init__(self, name: str, descr: str, group: ProductParamGroup, value: str, default: ProductParamDefault[str]):
        ProductParam.__init__(self, name, descr, ProductParamType.Hidden, group)
        self._value = value
        self._default = default

    @property
    def value(self) -> str:
        return self._value

    @value.setter
    def value(self, value: str) -> None:
        self._value = value

    @property
    def default(self) -> ProductParamDefault[str]:
        return self._default

    def get_string_value(self) -> str:
        return ParamStringConverter.to_string(self._value)
        
    def set_string_value(self, value: str) -> None:
        self._value = ParamStringConverter.from_string(value)


class ProductParamPid(ProductParam):
    def __init__(self, name: str, descr: str, group: ProductParamGroup, value: str, default: ProductParamDefault[str]):
        ProductParam.__init__(self, name, descr, ProductParamType.Pid, group)
        self._value = value
        self._default = default

    @property
    def value(self) -> str:
        return self._value

    @value.setter
    def value(self, value: str) -> None:
        self._value = value

    @property
    def default(self) -> ProductParamDefault[str]:
        return self._default

    def get_string_value(self) -> str:
        return ParamStringConverter.to_string(self._value)
        
    def set_string_value(self, value: str) -> None:
        self._value = ParamStringConverter.from_string(value)


class ProductParamBool(ProductParam):
    def __init__(self, name: str, descr: str, group: ProductParamGroup, value: bool, default: ProductParamDefault[bool]):
        ProductParam.__init__(self, name, descr, ProductParamType.Bool, group)
        self._value = value
        self._default = default
        self._child_param_list: List[ProductParam] = []

    @property
    def value(self) -> bool:
        return self._value

    @value.setter
    def value(self, value: bool) -> None:
        self._value = value

    @property
    def default(self) -> ProductParamDefault[bool]:
        return self._default
        
    @property
    def child_param_list(self) ->  List[ProductParam]:
        return self._child_param_list

    def get_string_value(self) -> str:
        return ParamBoolConverter.to_string(self._value)
        
    def set_string_value(self, value: str) -> None:
        self._value = ParamBoolConverter.from_string(value)


class ProductParamInt(ProductParam):
    def __init__(self, name: str, descr: str, group: ProductParamGroup, value: int, default: ProductParamDefault[int], param_range: ProductParamRange[int]):
        ProductParam.__init__(self, name, descr, ProductParamType.Int, group)
        self._value = value
        self._default = default
        self._param_range = param_range

    @property
    def value(self) -> int:
        return self._value

    @value.setter
    def value(self, value: int) -> None:
        self._value = value

    @property
    def default(self) -> ProductParamDefault[int]:
        return self._default

    @property
    def param_range(self) -> ProductParamRange[int]:
        return self._param_range

    def get_string_value(self) -> str:
        return ParamIntConverter.to_string(self._value)
        
    def set_string_value(self, value: str) -> None:
        self._value = ParamIntConverter.from_string(value)


class ProductParamFloat(ProductParam):
    def __init__(self, name: str, descr: str, group: ProductParamGroup, value: float, default: ProductParamDefault[float], param_range: ProductParamRange[float]):
        ProductParam.__init__(self, name, descr, ProductParamType.Float, group)
        self._value = value
        self._default = default
        self._param_range = param_range

    @property
    def value(self) -> float:
        return self._value

    @value.setter
    def value(self, value: float) -> None:
        self._value = value

    @property
    def default(self) -> ProductParamDefault[float]:
        return self._default

    @property
    def param_range(self) -> ProductParamRange[float]:
        return self._param_range

    def get_string_value(self) -> str:
        return ParamFloatConverter.to_string(self._value)
        
    def set_string_value(self, value: str) -> None:
        self._value = ParamFloatConverter.from_string(value)


class ProductParamString(ProductParam):
    def __init__(self, name: str, descr: str, group: ProductParamGroup, value: str, default: ProductParamDefault[str]):
        ProductParam.__init__(self, name, descr, ProductParamType.String, group)
        self._value = value
        self._default = default

    @property
    def value(self) -> str:
        return self._value

    @value.setter
    def value(self, value: str) -> None:
        self._value = value

    @property
    def default(self) -> ProductParamDefault[str]:
        return self._default

    def get_string_value(self) -> str:
        return ParamStringConverter.to_string(self._value)
        
    def set_string_value(self, value: str) -> None:
        self._value = ParamStringConverter.from_string(value)


class ProductParamEnum(ProductParam):
    def __init__(self, name: str, descr: str, group: ProductParamGroup, value: str,
                 default: ProductParamDefault[str], enum_values: List[str], multi: bool = False,
                 custom: bool = False):
        ProductParam.__init__(self, name, descr, ProductParamType.Enum, group)
        self._value = value
        self._default = default
        self._enum_values = enum_values
        self._multi = multi
        self._custom = custom

    @property
    def value(self) -> str:
        return self._value

    @value.setter
    def value(self, value: str) -> None:
        self._value = value

    @property
    def default(self) -> ProductParamDefault[str]:
        return self._default

    @property
    def enum_values(self) -> List[str]:
        return self._enum_values

    @property
    def multi(self) -> bool:
        return self._multi
        
    @multi.setter
    def multi(self, value: bool) -> None:
        self._multi = value

    @property
    def custom(self) -> bool:
        return self._custom
        
    @custom.setter
    def custom(self, value: bool) -> None:
        self._custom = value

    def get_string_value(self) -> str:
        return ParamStringConverter.to_string(self._value)
        
    def set_string_value(self, value: str) -> None:
        self._value = ParamStringConverter.from_string(value)
        
        
class ProductParamMapSizeRect(ProductParam):
    def __init__(self, name: str, descr: str, group: ProductParamGroup, value: MapSizeRect, default: ProductParamDefault[MapSizeRect], param_range: ProductParamRange[MapSizeRect]):
        ProductParam.__init__(self, name, descr, ProductParamType.MapSizeRect, group)
        self._value = value
        self._default = default
        self._param_range = param_range

    @property
    def value(self) -> MapSizeRect:
        return self._value

    @value.setter
    def value(self, value: MapSizeRect) -> None:
        self._value = value

    @property
    def default(self) -> ProductParamDefault[MapSizeRect]:
        return self._default

    @property
    def param_range(self) -> ProductParamRange[MapSizeRect]:
        return self._param_range

    def get_string_value(self) -> str:
        return ParamMapSizeRectConverter.to_string(self._value)
        
    def set_string_value(self, value: str) -> None:
        self._value = ParamMapSizeRectConverter.from_string(value)
        

class ProductParamPathDir(ProductParam):
    def __init__(self, name: str, descr: str, group: ProductParamGroup, value: str, default: ProductParamDefault[str]):
        ProductParam.__init__(self, name, descr, ProductParamType.PathDir, group)
        self._value = value
        self._default = default

    @property
    def value(self) -> str:
        return self._value

    @value.setter
    def value(self, value: str) -> None:
        self._value = value

    @property
    def default(self) -> ProductParamDefault[str]:
        return self._default

    def get_string_value(self) -> str:
        return ParamStringConverter.to_string(self._value)
        
    def set_string_value(self, value: str) -> None:
        self._value = ParamStringConverter.from_string(value)
        

class ProductParamPathFile(ProductParam):
    def __init__(self, name: str, descr: str, group: ProductParamGroup, value: str, default: ProductParamDefault[str]):
        ProductParam.__init__(self, name, descr, ProductParamType.PathFile, group)
        self._value = value
        self._default = default

    @property
    def value(self) -> str:
        return self._value

    @value.setter
    def value(self, value: str) -> None:
        self._value = value

    @property
    def default(self) -> ProductParamDefault[str]:
        return self._default

    def get_string_value(self) -> str:
        return ParamStringConverter.to_string(self._value)
        
    def set_string_value(self, value: str) -> None:
        self._value = ParamStringConverter.from_string(value)
        
        
class ProductParamCatchmentFile(ProductParam):
    def __init__(self, name: str, descr: str, group: ProductParamGroup, value: str, default: ProductParamDefault[str]):
        ProductParam.__init__(self, name, descr, ProductParamType.CatchmentFile, group)
        self._value = value
        self._default = default

    @property
    def value(self) -> str:
        return self._value

    @value.setter
    def value(self, value: str) -> None:
        self._value = value

    @property
    def default(self) -> ProductParamDefault[str]:
        return self._default

    def get_string_value(self) -> str:
        return ParamStringConverter.to_string(self._value)
        
    def set_string_value(self, value: str) -> None:
        self._value = ParamStringConverter.from_string(value)
        

class ProductParamRadarNetwork(ProductParam):
    def __init__(self, name: str, descr: str, group: ProductParamGroup, value: str, default: ProductParamDefault[str]):
        ProductParam.__init__(self, name, descr, ProductParamType.RadarNetwork, group)
        self._value = value
        self._default = default

    @property
    def value(self) -> str:
        return self._value

    @value.setter
    def value(self, value: str) -> None:
        self._value = value

    @property
    def default(self) -> ProductParamDefault[str]:
        return self._default

    def get_string_value(self) -> str:
        return ParamStringConverter.to_string(self._value)
        
    def set_string_value(self, value: str) -> None:
        self._value = ParamStringConverter.from_string(value)
        
        
class ProductInfo:
    def __init__(self, product_name: str=None, version: str=None, descr: str=None):
        self._product_name = product_name
        self._version = version
        self._descr = descr
        self._groups: List[ProductParamGroup] = []
        self._params: List[ProductParam] = []
            
    @property
    def product_name(self) -> str:
        return self._product_name

    @product_name.setter
    def product_name(self, value: str) -> None:
        self._product_name = value
        
    @property
    def version(self) -> str:
        return self._version

    @version.setter
    def version(self, value: str) -> None:
        self._version = value
        
    @property
    def description(self) -> str:
        return self._descr

    @description.setter
    def description(self, value: str) -> None:
        self._descr = value
        
    @property
    def groups(self) -> List[ProductParamGroup]:
        return self._groups
        
    @property
    def params(self) -> List[ProductParam]:
        return self._params
        
    def get_param(self, param_name: str) -> ProductParam:
        for param in self._params:
            if param.name == param_name:
                return param
                
        return None


class ProductParamInfo:
    def __init__(self, value=None, param_type: ProductParamType=ProductParamType.Undefined):
        self._value = value
        self._param_type = param_type
        
    @property
    def value(self):
        return self._value
    
    @value.setter
    def value(self, value) -> None:
        self._value = value
    
    @property
    def param_type(self) -> ProductParamType:
        return self._param_type
        
    @param_type.setter
    def param_type(self, value: ProductParamType) -> None:
        self._param_type = value


class ProdParamsBase(ABC):
    def __init__(self):
        self._map_param_info = {}
        self._product_info = ProductInfo()
        
    @property
    def map_info(self) -> Dict[str, ProductParamInfo]:
        return self._map_param_info
        
    @property
    def prod_info(self) -> ProductInfo:
        return self._product_info
    
    def inject(self, cmd_line_params: CmdLineParams) -> int:
        ret_counter = 0
        for name, value in cmd_line_params.params.items():
            info = self.get_param_info(name)
            if info.param_type != ProductParamType.Undefined:
                if info.param_type == ProductParamType.Hidden:
                    #special case for hidden parameters, handle in a hard-coded
                    #way the fields 'time' 'currentsweep' 'volumesweeps' and 'debug'
                    #all the other fields will be handled forcibly as strings
                    if name == "time":
                        info.value = ParamStringConverter.from_string(value)
                        ret_counter += 1
                    elif name in ["currentsweep", "volumesweeps", "debug"]:
                        info.value = ParamIntConverter.from_string(value)
                        ret_counter += 1
                    else:
                        info.value = ParamStringConverter.from_string(value)
                        ret_counter += 1
                elif info.param_type in [ProductParamType.Pid, ProductParamType.String, ProductParamType.Enum,
                        ProductParamType.PathDir, ProductParamType.PathFile,
                        ProductParamType.CatchmentFile, ProductParamType.RadarNetwork,
                        ProductParamType.RadarCfg]:
                    info.value = ParamStringConverter.from_string(value)
                    ret_counter += 1
                         
                elif info.param_type == ProductParamType.Bool:
                    info.value = ParamBoolConverter.from_string(value)
                    ret_counter += 1
                        
                elif info.param_type == ProductParamType.Int:
                    info.value = ParamIntConverter.from_string(value)
                    ret_counter += 1
                        
                elif info.param_type == ProductParamType.Float:
                    info.value = ParamFloatConverter.from_string(value)
                    ret_counter += 1
                        
                elif info.param_type == ProductParamType.MapSizeRect:
                    info.value = ParamMapSizeRectConverter.from_string(value)
                    ret_counter += 1
                        
        return ret_counter
        
    def get_params_cmd_line(self) -> str:
        ret = ""
        
        for name, info in self._map_param_info.items():
            if info.param_type != ProductParamType.Hidden:
                ret += f"{name}=\"{self.get_param_string(info)}\" "
                
        #remove the last space
        return ret[:-1]
    
    def get_xml_def(self) -> str:
        #create root node
        root_node = ET.Element("productdef")
        
        #create node info child of root
        info_node = ET.SubElement(root_node, "info")
        info_node.set("version", self._product_info.version)
        info_node.set("description", self._product_info.description)
        
        #create node params child of root
        params_node = ET.SubElement(root_node, "params")
        
        #create groups nodes children of params
        groups = self._product_info.groups
        params = self._product_info.params
        for group in groups:
            group_node = ET.SubElement(params_node, "group")
            group_node.set("name", group.name)
            
            #create param nodes children of group
            for param in params:
                #skip this param if does not belong to this group
                if param.group.name != group.name:
                    continue
                    
                #add this param to this group
                param_node = ET.SubElement(group_node, "param")
                param_node.set("name", param.name)
                param_node.set("type", self.get_param_type_string(param.param_type))
                param_node.set("default", param.get_string_value())
                
                #it it's an enum parameter add enum values
                #handle multi and custom parameters
                if param.param_type == ProductParamType.Enum:
                    #ppe = param
                    attr_multi = "1" if param.multi else "0"
                    param_node.set("multi", attr_multi)
                    attr_custom = "1" if param.custom else "0"
                    param_node.set("custom", attr_custom)
                    
                #if this parameter has range, add range
                param_min, param_max = self.get_param_range(param)
                if param_min is not None and param_max is not None:
                    param_node.set("range", f"{{{param_min}}}-{{{param_max}}}")
                    
                #add description attribute
                param_node.set("descr", param.descr)
                
                #it it's an enum parameter add enum values
                if param.param_type == ProductParamType.Enum:
                    enum_node = ET.SubElement(param_node, "values")
                    #ppe = param
                    for enum_value in param.enum_values:
                        value_node = ET.SubElement(enum_node, "value")
                        value_node.text = enum_value
        
        xml_str = minidom.parseString(ET.tostring(root_node, encoding='unicode')).toprettyxml(indent="    ")
        #xml_str = xml_str.replace('&lt;', '<').replace('&gt;', '>').replace('><', '>\n<').replace('\t', '    ')
        
        return xml_str
    
    def validate(self, cmd_line_params):
        for param_name, param_value in cmd_line_params.params.items():
            info = self.get_param_info(param_name)
            param = self._product_info.get_param(param_name)
            if info.param_type != ProductParamType.Undefined and param:
                if info.param_type == ProductParamType.Hidden:
                    #special case for hidden parameters, handle in a hard-coded
                    #way the fields 'time' currentsweep' 'volumesweeps' and 'debug'
                    #all the other fields will be handled forcibly as strings
                    if param_name == "time":
                        # TODO: implement custom validation for hidden field time
                        pass
                    elif param_name in ["currentsweep", "volumesweeps", "debug"]:
                        # TODO: implement custom validation for hidden fields
                        pass
                    else:
                        # TODO: implement custom validation for hidden fields
                        pass
                            
                elif info.param_type in [ProductParamType.Pid, ProductParamType.String,
                        ProductParamType.PathDir, ProductParamType.PathFile,
                        ProductParamType.CatchmentFile, ProductParamType.RadarNetwork,
                        ProductParamType.RadarCfg]:
                    #no validation, accept any value
                    pass
                        
                elif info.param_type == ProductParamType.Enum:
                    #implement validation for enum
                    #ppe = param.AsEnum()
                    #if it's not multi and it's not custom param_value
                    #must be equal to one of the possible values
                    if not param.multi and not param.custom:
                        if param_value not in param.enum_values:
                            acceptedValues = "|".join(param.enum_values)
                            raise ValueError(f"parameter '{param_name}': invalid value '{param_value}' for enum type. accepted values ({acceptedValues})")
                    
                elif info.param_type == ProductParamType.Bool:
                    #only '0' or '1' are accepted as valid values
                    if param_value not in ["0", "1"]:
                        raise ValueError(f"parameter '{param_name}': invalid value '{param_value}' for bool type. accepted values (0|1)")
                    
                elif info.param_type == ProductParamType.Int:
                    #validate string as valid integer number
                    if not re.match(r"^-?\d+$", param_value):
                        raise ValueError(f"parameter '{param_name}': invalid value '{param_value}' for int type. not an integer number")
                        
                    #validate range
                    value = int(param_value)
                    #paramInt = param.AsInt()
                    param_range = param.param_range
                    if param_range.has_range:
                        if not self.is_in_range(value, param_range.min_val, param_range.max_val):
                            str_min, str_max = self.get_param_range(param)
                            rangeValues = "{%s}-{%s}" % (str_min, str_max)
                            raise ValueError(f"parameter '{param_name}': invalid value '{param_value}' for range '{rangeValues}'")
                                
                elif info.param_type == ProductParamType.Float:
                        #validate string as valid float number
                        if not re.match(r"^-?\d+(\.\d*)?$", param_value):
                            raise ValueError(f"parameter '{param_name}': invalid value '{param_value}' for float type. not a floating point number")
                                
                        #validate range
                        value = float(param_value)
                        #paramFloat = param.AsFloat()
                        param_range = param.param_range
                        if param_range.has_range:
                            if not self.is_in_range(value, param_range.min_val, param_range.max_val):
                                str_min, str_max = self.get_param_range(param)
                                rangeValues = "{%s}-{%s}" % (str_min, str_max)
                                raise ValueError(f"parameter '{param_name}': invalid value '{param_value}' for range '{rangeValues}'")
                                
                elif info.param_type == ProductParamType.MapSizeRect:
                    #validate string as valid mapsizerect
                    value = ParamMapSizeRectConverter.from_string(param_value)
                    #paramRect = param
                    param_range = param.param_range
                    if param_range.has_range:
                        if not self.is_in_range(value.x_size, param_range.min_val.x_size,  param_range.max_val.x_size) or \
                            not self.is_in_range(value.y_size, param_range.min_val.y_size,  param_range.max_val.y_size) or \
                            not self.is_in_range(value.x_res, param_range.min_val.x_res,  param_range.max_val.x_res) or \
                            not self.is_in_range(value.y_res, param_range.min_val.y_res,  param_range.max_val.y_res):
                                str_min, str_max = self.get_param_range(param)
                                rangeValues = "{%s}-{%s}" % (str_min, str_max)
                                raise ValueError(f"parameter '{param_name}': invalid value '{param_value}' for range '{rangeValues}'")
        
    def get_param_info(self, param_name: str) -> ProductParamInfo:
        ret = self._map_param_info.get(param_name, ProductParamInfo(None, ProductParamType.Undefined))
        return self._map_param_info.get(param_name, ProductParamInfo(None, ProductParamType.Undefined))
        
    def get_param_string(self, param_info):
        if param_info.param_type in [ProductParamType.Pid, ProductParamType.String, ProductParamType.Enum,
                ProductParamType.PathDir, ProductParamType.PathFile,
                ProductParamType.CatchmentFile, ProductParamType.RadarNetwork, 
                ProductParamType.RadarCfg]:
            return ParamStringConverter.to_string(param_info.value)
        elif param_info.param_type == ProductParamType.Bool:
            return ParamBoolConverter.to_string(param_info.value)
        elif param_info.param_type == ProductParamType.Int:
            return ParamIntConverter.to_string(param_info.value)
        elif param_info.param_type == ProductParamType.Float:
            return ParamFloatConverter.to_string(param_info.value)
        elif param_info.param_type == ProductParamType.MapSizeRect:
            return ParamMapSizeRectConverter.to_string(param_info.value)
    
        else:
            return ""

    def get_param_type_string(self, param_type):
        if param_type == ProductParamType.Hidden:
            return "hidden"
        elif param_type == ProductParamType.Pid:
            return "pid"
        elif param_type == ProductParamType.String:
            return "string"
        elif param_type == ProductParamType.Enum:
            return "enum"
        elif param_type == ProductParamType.PathDir:
            return "pathdir"
        elif param_type == ProductParamType.PathFile:
            return "pathfile"
        elif param_type == ProductParamType.CatchmentFile:
            return "catchmentfile"
        elif param_type == ProductParamType.RadarNetwork:
            return "radarnetwork"
        elif param_type == ProductParamType.RadarCfg:
            return "radarcfg"
        elif param_type == ProductParamType.Bool:
            return "bool"
        elif param_type == ProductParamType.Int:
            return "int"
        elif param_type == ProductParamType.Float:
            return "float"
        elif param_type == ProductParamType.MapSizeRect:
            return "mapsizerect"
        else:
            return ""

    def get_param_range(self, param):
        param_type = param.param_type
    
        if param_type in [ProductParamType.Pid, ProductParamType.String, ProductParamType.Enum,
                ProductParamType.PathDir, ProductParamType.PathFile,
                ProductParamType.CatchmentFile, ProductParamType.RadarNetwork,
                ProductParamType.RadarCfg, ProductParamType.Bool]:
            return None, None
        elif param_type == ProductParamType.Int:
            param_range = param.param_range
            if param_range.has_range:
                min_val = ParamIntConverter.to_string(param_range.min_val)
                max_val = ParamIntConverter.to_string(param_range.max_val)
                return min_val, max_val
            else:
                return None, None
        elif param_type == ProductParamType.Float:
            param_range = param.param_range
            if param_range.has_range:
                min_val = ParamFloatConverter.to_string(param_range.min_val)
                max_val = ParamFloatConverter.to_string(param_range.max_val)
                return min_val, max_val
            else:
                return None, None
        elif param_type == ProductParamType.MapSizeRect:
            param_range = param.param_range
            if param_range.has_range:
                min_val = ParamMapSizeRectConverter.to_string(param_range.min_val)
                max_val = ParamMapSizeRectConverter.to_string(param_range.max_val)
                return min_val, max_val
            else:
                return None, None
        else:
            return None, None

    def is_in_range(self, value, val_min, val_max):
        return val_min <= value <= val_max
    
    @abstractmethod
    def init_param_info(self) -> None:
        """
        Subclasses must implement this method
        """
        pass
        
    @abstractmethod
    def init_product_info(self) -> None:
        """
        Subclasses must implement this method
        """
        pass
